#!/usr/bin/python3

from array import array

DETECTION_ALGORITHMS = [           # Method list for clap detection
    'threshold',
    'filter',
    'fft',
    'wavelet'
]

class SignalProcessor():
    """Describes all the signal processing algorithms and selector method.

    :param method: An object that contain the configuration and details of which algorithm to be used for processing the data received.
    :type method: class: `Munch`
    :var int alg_id: Stores the ID of the algorithm used
    :var algorithm: Stores the configuration of the algorithm used
    :vartype algorithm: class: `Munch`
    """

    def __init__(self, method):
        self.alg_id = DETECTION_ALGORITHMS.index(method.name)
        """**default:** ``0``

        Search for the algorithm name in the ``DETECTION_ALGORITHMS`` and store the id found
        """
        self.algorithm = method
        """The algorithm stored in this variable is used throughout the execution"""

    def findClap(self, data):
        """Based on the detection algorithm selected, the data is given to the selected algorithm

        :param data: Binary data received from microphone
        :type data: bytearray
        :return: `True` if clap is detected, `False` otherwise
        :rtype: bool
        """
        byte_stream = array('b', [0]) if data == None else data
        if self.alg_id == 1:
            return self.useFiter(byte_stream)
        elif self.alg_id == 2:
            return self.useFFT(byte_stream)
        elif self.alg_id == 3:
            return self.useWavelets(byte_stream)
        else:
            return self.useThreshold(byte_stream)

    def useFiter(self, byte_stream):
        """This algorithm is not implemented yet

        :param byte_stream: Binary stream of data received from microphone
        :type byte_stream: bytearray
        """
        # TODO
        return False

    def useFFT(self, byte_stream):
        """This algorithm is not implemented yet

        :param byte_stream: Binary stream of data received from microphone
        :type byte_stream: bytearray
        """
        # TODO
        return False

    def useWavelets(self, byte_stream):
        """This algorithm is not implemented yet

        :param byte_stream: Binary stream of data received from microphone
        :type byte_stream: bytearray
        """
        # TODO
        return False

    def useThreshold(self, byte_stream):
        """This algorithm uses a threshold value to detect claps.

        :param byte_stream: Binary stream of data received from microphone
        :type byte_stream: bytearray
        :return: When the maximum value of the integer array is greater than threshold, it returns `True` and `False` if otherwise
        :rtype: bool
        """
        max_value = max(array('h', byte_stream))
        return (True if max_value > self.algorithm.value else False)



#!/usr/bin/python3
import re
import os
import sys
from time import sleep
import _thread as thread
from array import array
import pyaudio
import statistics as stat

from ._settings import Settings
from ._processor import SignalProcessor


class Listener():
    """Describes methods which are called by user for the initialisation of the PyAudio module to stream microphone input.

    :param config: An object of :class:`Settings` which is used for configuring the module, defaults to None
    :type config: class: :class:`Settings`
    :param bool calibrate: If the flag is set, the chunk size is calibrated, defaults to True
    :var config: Store the `Settings` object
    :vartype config: class: `Settings`
    :var input: Store the `PyAudio` object
    :vartype input: class: `pyaudio.PyAudio`
    :var stream: Open a new input audio stream and store the stream object
    :vartype stream: class: `pyaudio.Stream`
    :var int claps: Store current value of claps counted and initially set to `0`
    :var lock: Store the thread lock
    :vartype lock: thread.lock
    :var processor: Store the `Processor` object
    :vartype processor: class: `SignalProcessor`
    """

    def __init__(self, config=None, calibrate=True):
        """Constructor method"""
        self.config = config or Settings()
        """If the :attr:`config` parameter is ``None``, an object of :class:`Settings` is assigned"""
        self.claps = 0
        """**default:** ``0``

        This class property contains the current count of claps detected"""
        self.lock = thread.allocate_lock()
        self.device = Device(self.config, calibrate)
        """Stores an :class:`Device` object which initialise :obj:`PyAudio` with calibration and manages the audio interface"""
        self.processor = SignalProcessor(method=self.config.method)
        """Initialised with an :class:`SignalProcessor` object using the signal processing method found inside :attr:`config`"""
        self.confirm()

    def clapWait(self, clap):
        """Start waiting for a small duration of time recursively until no more new claps are detected

        :param int clap: Number of claps found at the time of wait initialised
        """
        sleep(self.config.wait)
        if self.claps > clap:
            self.clapWait(self.claps)

    def listenClaps(self, threadName):
        """This method runs on a child thread with :attr:`lock` when :attr:`claps` equals ``1`` and reset the class property :attr:`claps` to ``0`` when execution is finished

        :param str threadName: Name of the child thread started
        """
        with self.lock:
            print("Waiting for claps...")
            self.clapWait(self.claps)
            action = 'on' + str(self.claps) + 'Claps'
            if action in self.config.actions:
                getattr(self.config, action)()
            print("You clapped", self.claps, "times.\n")
            self.claps = 0

    def __printInfo(self):
        print("Default Device\t\t:", self.device.defaultDevice['name'])
        print("Channels\t\t:", self.config.channels)
        print("Chunk size\t\t:", self.config.chunk_size, "bytes")
        print("Rate\t\t\t:", self.config.rate, "Hz")
        print("Clap wait\t\t:", self.config.wait, "sec")
        print("Algorithm selected\t:", self.config.method.name)
        print("Threshold Value\t\t:", self.config.method.value, "\n")

    def confirm(self):
        while True:
            self.__printInfo()
            ask = input('Confirm(c) / Recalibrate(r) / Manual(m) / Quit(q)?[c/r/m/q](default is c): ')
            if re.search(r'^[Cc]$|^$', ask):
                break
            elif re.search(r'^[Rr]$', ask):
                printf("\n")
                self.device.calibrateBufferSize()
            elif re.search(r'^[Mm]$', ask):
                t_value = input('Threshold value: ')
                self.config.method.value = int(t_value)
            elif re.search(r'^[Qq]$', ask):
                sys.exit(0)
            else:
                print("Invalid input. Try again.\n")

    def start(self):
        """When this method is called, the listener start reading binary data from stream and sreach for claps inside the chunks of data using :class:`SignalProcessor` until :attr:`Settings.exit` flag is ``True``

        :raises: ``KeyboardInterrupt``: If **Control + C** is pressed on keyboard
        """
        try:
            self.device.openStream()
            print("\nClap detection started")
            while not self.config.exit:
                try:
                    data = self.device.readData()
                except (OSError, IOError):
                    data = None
                if self.processor.findClap(data):
                    self.claps += 1
                if self.claps == 1 and not self.lock.locked():
                    thread.start_new_thread(
                        self.listenClaps, ("ListenClaps",))
        except(KeyboardInterrupt, SystemExit):
            pass
        self.stop()

    def stop(self):
        """When this method is called, the listener stop listening by closing the stream safely and terminating the connection"""
        print("\rExiting")
        self.device.closeStream()
        del self.device


class Device:

    def __init__(self, config, calibrate):
        self.config = config
        self.input = pyaudio.PyAudio()
        self.maxSamples = []
        os.system('clear')
        self.__setInputDevice()
        self.calibrateBufferSize(calibrate)
        # sys.exit(0)

    def __del__(self):
        self.input.terminate()

    def __setInputDevice(self):
        if self.input.get_host_api_count() < 1:
            print("No supported PortAudio Host APIs are found in your system")
            sys.exit(1)
        if self.input.get_device_count() < 1:
            print("No input audio device is found in your system")
            sys.exit(1)
        self.defaultDevice = self.input.get_default_input_device_info()
        self.config.channels = int(self.defaultDevice['maxInputChannels'])
        self.config.rate = int(self.defaultDevice['defaultSampleRate'])

    def readData(self):
        return self.stream.read(self.config.chunk_size)

    def openStream(self):
        self.stream = self.input.open(format=pyaudio.paInt16,
                                      channels=self.config.channels,
                                      rate=self.config.rate,
                                      input=True,
                                      frames_per_buffer=self.config.chunk_size)

    def closeStream(self):
        self.stream.stop_stream()
        self.stream.close()

    def calibrateBufferSize(self, enabled=True):
        if enabled:
            calibrated = False
            totalSamples = 400
            newChunkSize = self.config.chunk_size
            self.__printProgress(0, totalSamples)
            while not calibrated:
                try:
                    self.openStream()
                    for count in range(totalSamples):
                        data = self.stream.read(newChunkSize)
                        byte_stream = array('b', [0]) if data == None else data
                        maximum = max(array('h', byte_stream))
                        self.maxSamples.append(maximum)
                        self.__printProgress(count+1, totalSamples)
                    calibrated = True
                except(OSError):
                    if re.search(r'.+Input overflowed$', str(e)):
                        newChunkSize = int(newChunkSize / 2)
            self.config.chunk_size = newChunkSize
            self.closeStream()
            self.setThreshold()

    def setThreshold(self):
        maximum = max(self.maxSamples)
        median = stat.median_high(self.maxSamples)
        inter_value = median * 3.141592653589793
        value = stat.mean([inter_value,maximum])
        print(maximum,median,inter_value,value)
        self.config.method.value = int(value)

    def __printProgress(self, iteration, total):
        terminalSize = re.match("^\D*columns=(\d+), .*$",
                                str(os.get_terminal_size()))
        length = int(int(terminalSize.group(1)) - 24) if terminalSize else 75
        percent = ("{0:.1f}").format(100 * (iteration / float(total)))
        filledLength = int(length * iteration // total)
        bar = '█' * filledLength + '_' * (length - filledLength)
        if iteration != total:
            sys.stdout.write(f'Calibrating: [{bar}] {percent}%\r')
        else:
            sys.stdout.write(f'Calibrating: [{bar}] {percent}%\nCalibration complete\n\n')
        sys.stdout.flush()



"""A python package for clap detection

.. module:: pi-clap
   :platform: All
   :synopsis: A clap detection python module

.. moduleauthor:: Nikhil John <ceo@jwala.diamonds>

"""
from ._settings import Settings
from ._listener import Listener, Device
from ._processor import SignalProcessor

__project__ = 'pi-clap'
__version__ = '1.4.0'
__author__ = 'Nikhil John'
__all__ = ['Settings', 'Listener', 'Device', 'SignalProcessor']



"""

    ###################
    ##               ##
    ##    Pi-Clap    ##
    ##               ##
    ###################

Repo: https://github.com/nikhiljohn10/pi-clap
Author: Nikhil John
License: MIT
"""
from piclap import __version__ as VERSION

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pi-clap",
    version=VERSION,
    author="Nikhil John",
    author_email="ceo@jwala.diamonds",
    description="A python package for clap detection",
    url="https://pi-clap.nikz.in",
    long_description=long_description,
    long_description_content_type="text/markdown",
    download_url="https://github.com/nikhiljohn10/pi-clap",
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        'Environment :: Console',
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    platforms="any",
    py_modules=['piclap'],
    python_requires='>=3.6',
    install_requires=[
        "gpiozero==1.5.1",
        "pyaudio==0.2.11",
        "munch==2.5.0",
    ],
)



from munch import DefaultMunch as Objectify


class Settings:
    """This class describes all the configurations needed for the :class:`Listener` to work.

    :var boolean exit: Exit flag the determine the exit state of :class:`Listener`
    :var int rate: Bitrate at which input audio is streamed
    :var int channels: Number of audio channels used by :class:`Listener`
    :var int chunk_size: Frame count inside the audio buffer
    :var float wait: Clap wait in seconds

    :var method: The algorithm used for the detection of claps
    :vartype method: class: `Munch`
    :var actions: Collection of defined actions
    :vartype actions: list(str)
    """

    def __init__(self):
        """Constructor method"""
        self.exit = False
        """**default:** ``False``

        Exit flag
        """
        self.rate = 44100
        """**default:** ``44100``

        Number of audio samples collected in 1 second"""
        self.channels = 1
        """**default:** ``1`` (Mono Channel)

        Number of audio channel to listen"""
        self.chunk_size = 1024
        """**default:** ``1024``

        Number of frames in the input audio buffer"""
        self.wait = 0.5
        """**default:** ``0.5``

        Time duration to wait for claps to complete in :meth:`Listener.clapWait()`"""
        self.method = Objectify.fromDict({
            'name': 'threshold',
            'value': 512
        }, False)
        """**default:** :code:`{'name': 'threshold','value': 7000}`

        Detection method used for identifing claps"""
        self.actions = [m for m in dir(self) if m.startswith(
            'on') and m.endswith('Claps')]
        """When the class initialised, it collects all the actions defined inside this class as well as any classes where are derived with this class as base class

        **Condition:** *The method name defined should start with 'on' and end with 'Claps' with the clap count inbetween them.*
        """

    def on2Claps(self):
        """Action performed when 2 claps are detected."""
        print("Flashed light")

    def on3Claps(self):
        """Action performed when 3 claps are detected."""
        print("Toggled light")

    def on4Claps(self):
        """Action performed when 4 claps are detected. As default, :attr:`exit` flag is set to ``True`` if 4 claps are detected"""
        self.exit = True

    def updateMethod(self, method):
        """Update the method for detecting clap

        :param dict method: A dict type parameter which defines a clap detection method
        """
        self.method = Objectify.fromDict(method, False)
